import logging, json
import asyncio
from typing import Callable, Any, Union
from collections.abc import Iterable

from aiosow.autofill import autofill

ONS = {}


def on(variable_name: str, condition: Union[Callable, None] = None, singularize=False):
    """
    Decorator function that registers a function to be executed when a variable
    of specified name is perpetuated in memory.

    **Args**:
    - variable_name (str): Name of the event to listen to.
    - condition (Callable|None, optional): Callable object that takes the value
        of the event as input and returns a boolean indicating whether the
        registered function should be executed or not. Defaults to None.
    - singularize (bool, optional): Boolean indicating whether the value of the
        event should be singularized before being passed as an argument to the
        registered function. If True, the value must be iterable. Defaults to
        False.

    **Returns**:
    - The decorated function.

    **Example**:
    ```
    @on('pages')
    async def manage_pages(*args, memory):
        # code to send a welcome email to the user
    ```
    **note**: the decorated function is perpetuated and autofilled which means
    that it's result will be propegated to the memory and that it's arguments
    are autofilled based on it.
    """

    def decorator(function: Callable):
        if variable_name not in ONS:
            ONS[variable_name] = []
        ONS[variable_name].append((condition, function, singularize))
        return function

    return decorator


async def perpetuate(function: Callable, args: Any = [], memory: Any = {}) -> Any:
    """
    Asynchronously executes a function and perpetuates its effects in memory.

    **Args**:
    - function (Callable): The function to be executed.
    - args (Any, optional): Positional arguments to be passed to the function.
        Defaults to an empty list.
    - memory: The memory

    **Returns**:
    - The mutated keyword arguments of the executed function.

    **Example**:
    ```
    async def update_user_info(user_id, name=None, email=None):
        # code to update user info in the database
        return {'name': name, 'email': email}

    updated_values = await perpetuate(
        update_user_info, args=[123], memory={'name': 'John'}
    )
    # updated_values will be {'name': 'John', 'email': None}
    # (assuming the original value of email was None)
    ```
    """
    update = await autofill(function, args=args, memory=memory)
    if isinstance(update, dict):
        memory.update(update)
        logging.debug(
            "Mutation = %s", json.dumps(update, indent=4, default=lambda a: str(a))
        )
        # logging.debug('Memory = %s', json.dumps(memory, indent=4, default=lambda a: str(a)))
        for key, value in update.items():
            if key in ONS:
                for condition, func, singularize in ONS[key]:
                    if singularize:
                        if not isinstance(value, Iterable):
                            raise ValueError(
                                "Singularize received a non iterable value"
                            )
                        await asyncio.gather(
                            *[
                                autofill(func, args=[iterated], memory=memory)
                                for iterated in value.items()
                            ]
                            if isinstance(value, dict)
                            else [
                                autofill(func, args=[iterated], memory=memory)
                                for iterated in value
                            ]
                        )
                    else:
                        if (
                            condition
                            and await autofill(condition, args=[value], memory=memory)
                        ) or not condition:
                            await perpetuate(func, args=[value], memory=memory)
    return update



import pytest
from aiosow.routines import (
    routine,
    clear_routines,
    get_routines,
    consume_routines,
    infinite_generator,
)


@pytest.fixture(autouse=True)
def do_clear_routines():
    clear_routines()


@pytest.mark.asyncio
async def test_routine_decorator():
    @routine(0.2, perpetuate=False)
    def my_function():
        return {"test": "test"}

    memory = {}
    assert len(get_routines()) == 1
    assert get_routines()[0]["function"] == my_function
    assert get_routines()[0]["timeout"] == 0.2
    assert get_routines()[0]["perpetuate"] == False
    await consume_routines(memory=memory)
    assert memory.get("test", None) == None


@pytest.mark.asyncio
async def test_consume_routines_empty():
    def my_function():
        return {"test": "test"}

    routine(0.2)(my_function)
    memory = {}
    await consume_routines(memory=memory)
    assert memory["test"] == "test"


@pytest.mark.asyncio
async def test_routines_should_raise_if_raise_true():
    def my_function():
        raise ValueError

    routine(0.1)(my_function)
    assert len(get_routines()) == 1
    memory = {"raise": True}
    with pytest.raises(Exception):
        await consume_routines(memory=memory)
    memory = {"raise": False}
    await consume_routines(memory=memory)


@pytest.mark.asyncio
async def test_routines_should_be_removed_with_negative_frequency():
    def my_function():
        raise ValueError

    memory = {}
    routine(-0.1)(my_function)
    assert len(get_routines()) == 1
    await consume_routines(memory=memory)
    assert len(get_routines()) == 0


async def test_infinite_generator():
    async def my_condition():
        return True

    async def my_function():
        yield 1
        yield 2

    generator_consumer = infinite_generator(my_condition)(my_function)

    assert await generator_consumer() == 1
    assert await generator_consumer() == 2
    assert await generator_consumer() == 1



import pytest

from aiosow.perpetuate import perpetuate, on


@pytest.mark.asyncio
async def test_perpetuate():
    mem = {"a": 1}

    def test_function(a):
        return a + 1

    def test_function2(a):
        return {"result": a + 1}

    assert await perpetuate(test_function, memory=mem) == 2
    assert await perpetuate(test_function2, memory=mem) == {"result": 2}

    assert mem["result"] == 2


@pytest.mark.asyncio
async def test_on(mocker):
    mem = {"a": 1, "b": []}
    mock = mocker.Mock()
    mockb = mocker.Mock()
    on("a")(mock)
    on("b", singularize=True)(mockb)
    assert await perpetuate(lambda: {"a": 2}, memory=mem)
    mock.assert_called_once()
    assert await perpetuate(lambda: {"b": [1, 2]}, memory=mem)
    assert mockb.call_count == 2
    with pytest.raises(ValueError):
        await perpetuate(lambda: {"b": 1}, memory=mem)



import logging, asyncio
from typing import Dict, List, Callable
from aiosow.perpetuate import perpetuate

SETUP_FUNCTIONS: List = []


def clear_setups():  # pragma: no cover
    global SETUP_FUNCTIONS
    SETUP_FUNCTIONS = []


def setup(func: Callable) -> Callable:
    """
    Decorator to add a function to the list of initialization functions.

    **Args**:
    - func (Callable): Function to add to the list of initialization functions.

    **Returns**:
    - func (Callable): The same function, unchanged.
    """
    if func in SETUP_FUNCTIONS:
        logging.debug("duplicate setup declaration of %s is ignored", func.__name__)
    else:
        logging.info("+ setup %s", func.__name__)
        SETUP_FUNCTIONS.append(func)
    return func


async def initialize(memory: Dict) -> List[asyncio.Task]:
    """
    Function that runs all initialization functions added to the list.

    Args:
        - app (web.Application): The aiohttp application.
        - mem (Dict): The mem dictionary.
    """
    logging.debug("initialize with %s", [f"{fn.__name__}" for fn in SETUP_FUNCTIONS])
    tasks = []
    for setup_func in SETUP_FUNCTIONS:
        logging.debug("setup(%s)", setup_func.__name__)
        result = await perpetuate(setup_func, memory=memory)
        logging.debug("%s() => %s", setup_func.__name__, result)
        if asyncio.iscoroutine(result):
            tasks.append(result)
        logging.debug(f"{setup_func.__module__}.{setup_func.__name__} : ok")
    logging.debug("initialize is done")
    return tasks


__all__ = ["setup"]



#! env python3.11

import logging, asyncio, importlib, argparse, sys
from aiosow.setup import initialize
from aiosow.options import options
from aiosow.routines import spawn_routine_consumer


def load_composition(composition=None):
    debug = (
        "-d" in sys.argv
        or "--debug" in sys.argv
        or any(arg.find("-d") != -1 or arg.find("--debug") != -1 for arg in sys.argv)
    )
    if not "-h" in sys.argv:
        logging.basicConfig(
            level=logging.DEBUG if debug else logging.INFO,
            format="[%(asctime)s][%(levelname)s] \t%(message)s",
            datefmt="%H:%M:%S",
        )
    parser = argparse.ArgumentParser()
    if not composition:
        parser.add_argument("composition", help="composition to run")
    # parser.add_argument(
    #    "-c", "--config", help="Path to configuration file", default="", type=str
    # )
    parser.add_argument(
        "-d", "--debug", default=False, action="store_true", help="Debug mode"
    )
    parser.add_argument(
        "-la",
        "--log_autofill",
        default=False,
        action="store_true",
        help="Log.info autofilled functions",
    )
    parser.add_argument(
        "--raise",
        default=False,
        action="store_true",
        help="Routines will raise errors and stop the execution",
    )
    try:
        if not composition:
            composition = sys.argv[1]
        composition = (
            f"{composition}.bindings" if not ".bindings" in composition else composition
        )
        importlib.import_module(composition)
    except Exception as e:
        logging.error("An error occured opening the composition")
        raise (e)
    for module_name in sys.modules.keys():
        if "bindings" in module_name and "aiosow" not in module_name:
            logging.info("+ composition %s", module_name)
    logging.info("--------------------------------")
    logging.info("-------------START--------------")
    logging.info("--------------------------------")
    for args, kwargs in options():
        parser.add_argument(*args, **kwargs)
    return vars(parser.parse_args())


def run(composition=None):
    memory = load_composition(composition=composition)
    logging.debug(memory)
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    tasks = loop.run_until_complete(initialize(memory))
    memory["running"] = True
    logging.info("--------------------------------")
    logging.info("----------SETUP--DONE-----------")
    logging.info("--------------------------------")
    consumer = loop.run_until_complete(spawn_routine_consumer(memory))
    tasks = tasks + [consumer]
    # setups can return a task which is ran here
    # this allows setups to start tasks and still have them complete
    loop.run_until_complete(asyncio.gather(*tasks))
    if memory["run_forever"]:
        loop.run_forever()


if __name__ == "__main__":
    run()

__all__ = []



from setuptools import setup, find_packages


with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="aiosow",
    version="0.1.3",
    packages=find_packages(include=["aiosow"]),
    description="An event-based framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    entry_points={
        "console_scripts": [
            "aiosow = aiosow.command:run",
        ],
    },
    install_requires=[],
    extras_require={
        "dev": [
            "pytest",
            "pytest-mock",
            "pytest-asyncio",
            "pytest-coverage",
        ],
        "doc": ["pdoc"],
    },
)



"""
When writing your own compositions you may want to add additionnal options to the
command-line.

You do it using the option function.

**Example:**
```
option('keyword', default='foo', help='Keyword to use')
```

Any option defined is propagated into the memory, in fact the initial memory is
a copy of the parsed arguments.
"""

OPTIONS = []


def option(*args, **kwargs):  # pragma: no cover
    """
    Register an argparse option to be available using the command-line.

    **args**:
    - name: str -> the name of the option (will be used as f'--{name}')
    - kwargs: dict -> the options used by [argparse](https://docs.python.org/3/library/argparse.html)
    """
    global OPTIONS
    OPTIONS.append((args, kwargs))


def options():  # pragma: no cover
    global OPTIONS
    return OPTIONS


__all__ = ["option"]
